//FragmentShader
#version 430 core

struct TextureStruct {
    sampler2D diffuse;
    sampler2D specular;
    sampler2D normal;
    sampler2D dudv;
}; 

struct Material {
    float diffuse;
    float specular;
    float shininess;
    float ambient;
}; 

struct DirLight {
    vec3 direction;
    vec3 lightColor;
};

struct PointLight {
    vec3 position;
    
    float constant;
    float linear;
    float quadratic;
	
    vec3 lightColor;
};

struct SpotLight {
    vec3 position;
    vec3 direction;
    float cutOff;
    float outerCutOff;
  
    float constant;
    float linear;
    float quadratic;
  
    vec3 lightColor;     
};

#define NR_POINT_LIGHTS 20
#define NR_DIRECTIONAL_LIGHTS 1

uniform DirLight dirLight[NR_DIRECTIONAL_LIGHTS];
uniform PointLight pointLights[NR_POINT_LIGHTS];
uniform SpotLight spotLight;
uniform vec4 material; // The material's coefficients ([0] = ambient, [1] = diffuse, [2] = specular, [3] = shininess)
uniform TextureStruct textureStruct;
layout(location = 10) uniform int shaderVersion;
layout(location = 11) uniform int normalUse;
layout(location = 12) uniform int sky;
layout(location = 13) uniform float brightness;
layout(location = 14) uniform int onlyBlack;
layout(location = 15) uniform int pointlightsToRender;
layout(location = 16) uniform int transparency;
layout(location = 17) uniform int dudv;
layout(location = 18) uniform float moveFactor;
layout(location = 35) uniform samplerCube skybox;

const float waveStrength = 1.0;

in vec3 FragPos; 
in vec3 Normal;
in vec3 viewDir;
in vec2 TexCoords;
in vec2 textureCoords;
in mat3 TBN;

// function prototypes
vec3 CalcDirLight(DirLight light, vec3 normal, vec3 reflectColor, vec3 texColor, vec3 texSpec);
vec3 CalcPointLight(PointLight light, vec3 normal, vec3 reflectColor, vec3 texColor, vec3 texSpec);
vec3 CalcSpotLight(SpotLight light, vec3 normal, vec3 reflectColor, vec3 texColor, vec3 texSpec);


out vec4 FragColor;

void main()
{ 
    if (onlyBlack == 1){
            FragColor = vec4(0.0f, 0.0f, 0.0f, 1.0);
    } else {
        if (sky == 1){
    	    vec3 skyColor = vec3(texture(skybox, -viewDir).rgb);
            FragColor = vec4(skyColor * brightness, 1.0);
        } else {
            // calculate the normals from input and from a normal map, decide with the normalUse which normals are used
            vec3 normInput = normalize(Normal);
            vec3 normTextIn = normalize((2 * vec3(texture(textureStruct.normal, TexCoords).rgb)) - 1);
            normTextIn = normalize(TBN * normTextIn);
            vec3 normCalc[2] = { normInput, normTextIn };

            vec3 norm = normalize(normCalc[normalUse]);

            // calculate coordinate for the skybox
            vec3 reflectionCoord = reflect(-viewDir, normalize(norm));

            //texture color
            vec3 texColor = vec3(texture(textureStruct.diffuse, TexCoords).rgb);

            //texture specular value
            vec3 texSpec = vec3(texture(textureStruct.specular, TexCoords).rgb);

            if (dudv == 1){
                vec3 norm1 = normalize((texture(textureStruct.normal, vec2(TexCoords.x + moveFactor, TexCoords.y)).rgb * 2.0 - 1.0) * waveStrength);
                vec3 norm2 = normalize((texture(textureStruct.normal, vec2(-TexCoords.x + moveFactor, TexCoords.y + moveFactor)).rgb * 2.0 - 1.0) * waveStrength);
                norm = normalize(TBN * normalize(norm1 + norm2));

                reflectionCoord = reflect(-viewDir, normalize(norm));

                texColor = vec3(0.8, 0.8, 0.8);
            }

            //skybox color
	        vec3 reflectColor = vec3(texture(skybox, reflectionCoord).rgb);

            //initial result with ambient color
            vec3 result = texColor * material[0]; // ambient

            // == =====================================================
            // The lighting is set up in 3 phases: directional, point lights and an optional flashlight
            // For each phase, a calculate function is defined that calculates the corresponding color
            // per lamp. In the main() function we take all the calculated colors and sum them up for
            // this fragment's final color.
            // == =====================================================
	            
            // phase 1: directional lighting
	        result += CalcDirLight(dirLight[0], norm, reflectColor, texColor, texSpec);
	        for(int i = 1; i < NR_DIRECTIONAL_LIGHTS; i++)
	        	result += CalcDirLight(dirLight[i], norm, reflectColor, texColor, texSpec);
            // phase 2: point lights
            for(int i = 0; i < pointlightsToRender; i++)
                result += CalcPointLight(pointLights[i], norm, reflectColor, texColor, texSpec);    
            // phase 3: spot light
            // result += CalcSpotLight(spotLight, norm, FragPos, viewDir, reflectColor, texColor, texSpec); 
            if (transparency == 1){
                FragColor = vec4(result * brightness, 0.5);
            } else {
                FragColor = vec4(result * brightness, 1.0);
            }
            
        }
    }
} 


// calculates the color when using a directional light.
vec3 CalcDirLight(DirLight light, vec3 normal, vec3 reflectColor, vec3 texColor, vec3 texSpec)
{
	vec3 lightDir = normalize(-light.direction);
	vec3 reflectDir = reflect(-lightDir, normal);

    // diffuse shading
    float diff = max(dot(normal, lightDir), 0.0f);

    vec3 normalDiff = light.lightColor * diff * material[1] * texColor;
	vec3 environmentDiff = light.lightColor * diff * material[1] * texColor + (reflectColor * 0.3f);
	vec3 diffCalc[4] = {
			normalDiff,
			normalDiff,
			environmentDiff,
			environmentDiff
	};

    // specular shading
    float spec = pow(max(dot(viewDir, reflectDir), 0.0f), material[3]);

    vec3 texturSpec = light.lightColor * spec * texSpec;
	vec3 normalSpec = light.lightColor * spec * material[2];
	vec3 environmentTexturSpec = light.lightColor * spec * texSpec;
	vec3 environmentNormalSpec = light.lightColor * spec * material[2];
	vec3 specCalc[4] = {
			normalSpec,
			texturSpec,
			environmentNormalSpec,
			environmentTexturSpec
	};

    // combine results
    vec3 diffuse = diffCalc[shaderVersion];
	vec3 specular = specCalc[shaderVersion];
    
    return (diffuse + specular);
}

// calculates the color when using a point light.
vec3 CalcPointLight(PointLight light, vec3 normal, vec3 reflectColor, vec3 texColor, vec3 texSpec)
{
    vec3 lightDir = normalize(light.position - FragPos);

    // diffuse shading
    float diff = max(dot(normal, lightDir), 0.0f);

    vec3 normalDiff = light.lightColor * diff * material[1] * texColor;
	vec3 environmentDiff = light.lightColor * diff * material[1] * texColor + (reflectColor * 0.3f);
	vec3 diffCalc[4] = {
			normalDiff,
			normalDiff,
			environmentDiff,
			environmentDiff
	};

    // specular shading
    vec3 reflectDir = reflect(-lightDir, normal);
    float spec = pow(max(dot(viewDir, reflectDir), 0.0f), material[3]);

    vec3 texturSpec = light.lightColor * spec * texSpec;
	vec3 normalSpec = light.lightColor * spec * material[2];
	vec3 environmentTexturSpec = light.lightColor * spec * texSpec;
	vec3 environmentNormalSpec = light.lightColor * spec * material[2];
	vec3 specCalc[4] = {
			normalSpec,
			texturSpec,
			environmentNormalSpec,
			environmentTexturSpec
	};

    // attenuation
    float lightDistance = length(light.position - FragPos) / 8.0f;
    float attenuation = 1.0f / (light.constant + light.linear * lightDistance + light.quadratic * lightDistance * lightDistance);   
    
    // combine results
    vec3 diffuse = diffCalc[shaderVersion];
	vec3 specular = specCalc[shaderVersion];

    return (diffuse + specular) * attenuation;
}

// calculates the color when using a spot light.
vec3 CalcSpotLight(SpotLight light, vec3 normal, vec3 reflectColor, vec3 texColor, vec3 texSpec)
{
    vec3 lightDir = normalize(light.position - FragPos);

    // diffuse shading
    float diff = max(dot(normal, lightDir), 0.0f);

    vec3 normalDiff = light.lightColor * diff * material[1] * texColor;
	vec3 environmentDiff = light.lightColor * diff * material[1] * texColor + (reflectColor * 0.3f);
	vec3 diffCalc[4] = {
			normalDiff,
			normalDiff,
			environmentDiff,
			environmentDiff
	};

    // specular shading
    vec3 reflectDir = reflect(-lightDir, normal);
    float spec = pow(max(dot(viewDir, reflectDir), 0.0f), material[3]);

    vec3 texturSpec = light.lightColor * spec * texSpec;
	vec3 normalSpec = light.lightColor * spec * material[2];
	vec3 environmentTexturSpec = light.lightColor * spec * texSpec;
	vec3 environmentNormalSpec = light.lightColor * spec * material[2];
	vec3 specCalc[4] = {
			normalSpec,
			texturSpec,
			environmentNormalSpec,
			environmentTexturSpec
	};

    // attenuation
    float distance = length(light.position - FragPos);
    float attenuation = 1.0f / (light.constant + light.linear * distance + light.quadratic * (distance * distance));    

    // spotlight intensity
    float theta = dot(lightDir, normalize(-light.direction)); 
    float epsilon = light.cutOff - light.outerCutOff;
    float intensity = clamp((theta - light.outerCutOff) / epsilon, 0.0f, 1.0f);

    // combine results
    vec3 diffuse = diffCalc[shaderVersion];
	vec3 specular = specCalc[shaderVersion];

    return (diffuse + specular) * attenuation * intensity;
}